#include "actions.h"
#include <engine_engine_number_9/baza_throw.h>
#include "str_fields.h"
#include <sqlite3.h>
#include <engine_engine_number_9/os_utils.h>
#include "find_db.h"
#include <unistd.h>
#include <assert.h>
#include "sqlite3_wrapper.h"

namespace iu9cawebchat {
    void initialize_website(const json::JSON& config, const std::string& root_pw) {
        printf("Initialization...\n");
        een9_ASSERT(check_password(root_pw), "Bad root password");
        std::string db_path;
        int ret;
        ret = find_db_sqlite_file_path(config, db_path);
        een9_ASSERT(ret == 0, "Invalid settings[\"database\"] field");
        if (een9::isRegularFile(db_path)) {
            // todo: plaese, don't do this
            ret = unlink(db_path.c_str());
            een9_ASSERT_pl(ret == 0);
        }
        een9_ASSERT(!een9::isRegularFile(db_path), "Database file exists prior to initialization. "
            "Can't preceed withut harming existing data");
        SqliteConnection conn(db_path.c_str());
        assert(sqlite3_errcode(conn.hand) == SQLITE_OK);
        /* Role of memeber of chat:
         * 1 - admin
         * 2 - regular
         * 3 - read-only member
         * 4 - deleted (no longer a member)
         *
         * If user.id is 0, it is a root user
         * If chat.lastMsgId is NULL, chat is empty
         * If message.previous is NULL, this message is first in it's chat
         */
        sqlite_nooutput(conn.hand, "PRAGMA foreign_keys = true");
        sqlite_nooutput(conn.hand, "BEGIN");
        sqlite_nooutput(conn.hand, "CREATE TABLE `nickname` (`it` TEXT PRIMARY KEY NOT NULL) WITHOUT ROWID");
        sqlite_nooutput(conn.hand, "CREATE TABLE `user` ("
                                         "`id` INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,"
                                         "`nickname` TEXT UNIQUE REFERENCES `nickname` NOT NULL,"
                                         "`name` TEXT NOT NULL,"
                                         "`chatList_HistoryId` INTEGER NOT NULL,"
                                         "`password` TEXT NOT NULL"
                                         ")");
        sqlite_nooutput(conn.hand, "CREATE TABLE `chat` ("
                                         "`id` INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,"
                                         "`nickname` TEXT UNIQUE REFERENCES `nickname` NOT NULL,"
                                         "`name` TEXT NOT NULL,"
                                         "`it_HistoryId` INTEGER NOT NULL,"
                                         "`lastMsgId` INTEGER REFERENCES `message`"
                                         ")");
        sqlite_nooutput(conn.hand, "CREATE TABLE `user_chat_membership` ("
                                         "`userId` INTEGER REFERENCES `user` NOT NULL,"
                                         "`chatId` INTEGER REFERENCES `chat` NOT NULL,"
                                         "`user_chatList_IncHistoryId` INTEGER NOT NULL,"
                                         "`chat_IncHistoryId` INTEGER NOT NULL,"
                                         "`role` INTEGER NOT NULL"
                                         ")");
        sqlite_nooutput(conn.hand, "CREATE TABLE `message` ("
                                         "`id` INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,"
                                         "`chatId` INTEGER REFERENCES `chat` NOT NULL,"
                                         "`previous` INTEGER REFERENCES `message`,"
                                         "`senderUserId` INTEGER REFERENCES `user` NOT NULL,"
                                         "`exists` BOOLEAN NOT NULL,"
                                         "`isSystem` BOOLEAN NOT NULL,"
                                         "`text` TEXT NOT NULL,"
                                         "`chat_IncHistoryId` INTEGER NOT NULL"
                                         ")");
        sqlite_nooutput(conn.hand, "INSERT INTO `nickname` VALUES (?1)", {}, {{1, "root"}});
        sqlite_nooutput(conn.hand, "INSERT INTO `user` (`id`, `nickname`, `name`, `chatList_HistoryId`, `password`) VALUES "
                                         "(0, ?1, ?2, 0, ?3)", {},
                                         {{1, "root"}, {2, "Rootov Root Rootovich"}, {3, root_pw}});
        sqlite_nooutput(conn.hand, "END");
    }
}
